import numpy as np

class Graph:
        
    def __init__(self, 
                 k_hop, mode):
        
        """
        inputs:
            mode: "hand", "pose", "pose_hand"
        """
        
        self.k_hop = k_hop
        self.mode = mode
        
        self.create_vertex()
        self.create_edge()
        self.create_adjacent_matrix()
        self.create_D()
        self.create_k_hop_matrix()
        
    def create_vertex(self):
        self.pose_num_vertices = 33
        self.hand_num_vertices = 21
        if self.mode == "hand":
            self.num_vertices = self.hand_num_vertices * 2
        elif self.mode == "pose":
            self.num_vertices = self.pose_num_vertices
        elif self.mode == "pose_hand":
            self.num_vertices = self.pose_num_vertices + \
                                self.hand_num_vertices * 2
            
            
        
    def create_edge(self):
        """
        The pose edges are defined by https://google.github.io/mediapipe/solutions/pose.html
        The hand edges are defined by https://google.github.io/mediapipe/solutions/hands.html
        """

        self_edges = [(v, v) for v in range(self.num_vertices)]
        neighbor_edges = [(0, 1), (0, 4), (1, 2), (2, 3),
                         (3, 7), (4, 5), (5, 6), (6, 8), 
                         (9, 10), (11, 12), (11, 13), (11, 23), (12, 14),
                         (12, 24), (13, 15), (14, 16), (15, 17), (15, 19),
                         (15, 21), (16, 18), (16, 20), (16, 22), (17, 19),
                         (18, 20), (23, 24), (23, 25), (24, 26), (25, 27),
                         (26, 28), (27, 29), (27, 31), (28, 30), (28, 32), 
                         (29, 31), (30, 32), 
                        ]
        
        hand_edges = [
            (0, 1), (0, 5), (0, 9), (0, 13), (0, 17),
            (1, 2), (2, 3), (3, 4), (5, 6), (5, 9), 
            (6, 7), (7, 8), (9, 10), (9, 13), (10, 11),
            (11, 12), (13, 14), (13, 17), (14, 15), (15, 16),
            (17, 18), (18, 19), (19, 20)
        ]
        
        if self.mode == "hand":
            left_hand_edges = hand_edges.copy()
            right_hand_edges = [(i+self.hand_num_vertices, j+self.hand_num_vertices) \
                                 for i, j in hand_edges]
            self.edges = self_edges + left_hand_edges + right_hand_edges
        elif self.mode == "pose":
            self.edges = self_edges + neighbor_edges
        elif self.mode == "pose_hand":
            left_hand_edges = [(i+self.pose_num_vertices, j+self.pose_num_vertices) \
                                 for i, j in hand_edges]
            right_hand_edges = [(i+self.hand_num_vertices, j+self.hand_num_vertices) \
                                 for i, j in left_hand_edges]
            self.edges = self_edges + neighbor_edges + left_hand_edges + right_hand_edges
        
    def create_adjacent_matrix(self):
        self.A = np.zeros((self.num_vertices, self.num_vertices))
        for i, j in self.edges:
            self.A[i, j] = 1
            self.A[j, i] = 1
        
    def create_D(self):
        sum_row = self.A.sum(axis=1)
        self.D = np.diag(sum_row)
        
    def normalize_A(self, A):
        return np.linalg.inv(self.D) @ A
    
    def create_k_hop_matrix(self):
        
        As = []
        
        for hop in range(1, self.k_hop + 1):
            As.append(self.normalize_A(np.power(self.A, hop)))
        
        self.A_k = np.stack(As, axis=0)
        


import mediapipe as mp
import pandas as pd
from google.protobuf.json_format import MessageToDict
from ..utils.key_points import W_LIST_POSE, W2I_POSE, W_LIST_LEFT_HAND, W2I_LEFT_HAND, W_LIST_RIGHT_HAND, W2I_RIGHT_HAND

class HolisticTracker:
    
    def __init__(self):
        self.mp_holistic = mp.solutions.holistic
        self.holistic = self.mp_holistic.Holistic(min_detection_confidence=0.5,
                                               min_tracking_confidence=0.5)
        self.mp_drawing = mp.solutions.drawing_utils
        self.result_hands = None
        self.target_columns = [key + "_x" for key in W_LIST_RIGHT_HAND] + [key + "_y" for key in W_LIST_RIGHT_HAND]
        self.target_columns += [key + "_x" for key in W_LIST_LEFT_HAND] + [key + "_y" for key in W_LIST_LEFT_HAND]
        self.target_columns += [key + "_x" for key in W_LIST_POSE] + [key + "_y" for key in W_LIST_POSE]

    def __call__(self, image, ts):
        """
        input:
            image: image
        output:
            output: hand_skeleton
        """

        self.results = self.holistic.process(image)
        keypoints_dict = self.pred_to_dict(self.results, ts)
        skeleton = self.dict_to_data(keypoints_dict)

        return skeleton

    def dict_to_data(self, keypoints_dict):
        df = pd.DataFrame(keypoints_dict, index=[0])               

        return df

    def pred_to_dict(self, results, ts):
        
        cp_dict = {}
        cp_dict_list = []

        if results and results.pose_landmarks:
            cp_dict_list.append(
                self.landmarks_to_dict(results.pose_landmarks.landmark, W2I_POSE, ts)
            )
        else:
            cp_dict_list.append(
                self.landmarks_to_dict_dummy(W2I_POSE, ts)
            )

        if results and results.left_hand_landmarks:
            cp_dict_list.append(
                self.landmarks_to_dict(results.left_hand_landmarks.landmark, W2I_LEFT_HAND, ts)
            )
        else:
            cp_dict_list.append(
                self.landmarks_to_dict_dummy(W2I_LEFT_HAND, ts)
            )

        if results and results.right_hand_landmarks:
            cp_dict_list.append(
                self.landmarks_to_dict(results.right_hand_landmarks.landmark, W2I_RIGHT_HAND, ts)
            )
        else:
            cp_dict_list.append(
                self.landmarks_to_dict_dummy(W2I_RIGHT_HAND, ts)
            )

        for cp in cp_dict_list:
            cp_dict.update(cp)

        return cp_dict

    def landmarks_to_dict_dummy(self, w2i, ts):
 
        cp_dict = {}

        for w in w2i:
            cp_dict[w + "_x"] = 0.0
            cp_dict[w + "_y"] = 0.0
            cp_dict[w + "_v"] = 0.0

        cp_dict["ts"] = ts
        
        return cp_dict

    def landmarks_to_dict(self, landmarks, w2i, ts):

        cp_dict = {}

        for w in w2i:
            cp_dict[w + "_x"] = landmarks[w2i[w]].x
            cp_dict[w + "_y"] = -landmarks[w2i[w]].y + 1
            cp_dict[w + "_v"] = landmarks[w2i[w]].visibility

        cp_dict["ts"] = ts
        
        return cp_dict

    def draw_landmarks(self, image):
        
        if self.results:
            self.mp_drawing.draw_landmarks(
                image, self.results.left_hand_landmarks, self.mp_holistic.HAND_CONNECTIONS)
            self.mp_drawing.draw_landmarks(
                image, self.results.right_hand_landmarks, self.mp_holistic.HAND_CONNECTIONS)
            self.mp_drawing.draw_landmarks(
                image, self.results.pose_landmarks, self.mp_holistic.POSE_CONNECTIONS)

        return image


import threading
import time
import mouse
import keyboard
import pandas as pd
import pathlib
import re
import os
import numpy as np
import csv
import traceback

from collections import defaultdict, deque
from PIL import Image

from ..module import Module, KeyControl
from ..detector import RobloxLiftGameActionDetector, MouseControlGestureDetector

class HAC:

    saved_images = []    # images to data
    images = deque()     # captured images
    tss = deque()        # timestamps
    controls = deque()   # controls need to be executed
    modules = {}         # available modules of this controller
    module = None        # current module
    buffer_size = 10

    def __init__(self, holistic_tracker):
        """
        Human Action Controller (HAC)
        A Human action controller needs a holistic tracker to extract a skeleton from an image.

        inputs:
            holistic_tracker: a tracker to extract a skeleton from an image
                              take hac.tracker.holistic_tracker as a reference.
        """
        self.holistic_tracker = holistic_tracker

    def add_module(self, module_name):
        """
        Add an module to HAC.
        
        inputs: 
            module_name: a module contains a set of actions which can be detected by a detector.
        """
        if module_name == "mouse":
            detector = MouseControlGestureDetector()
        if module_name == "roblox_lift_game":
            detector = RobloxLiftGameActionDetector()

        module = Module(detector)
        self.modules[module_name] = module

        return module
        
    def set_init_module(self, module):
        """
        Set a default module. Check hac.module to know more about modules
        """

        self.module = module

    def buffer_resize(self):

        if len(self.images) > self.buffer_size:
            self.images.popleft()

        if len(self.tss) > self.buffer_size:
            self.tss.popleft()

        if len(self.controls) > self.buffer_size:
            self.controls.popleft()

    def update(self, image, ts, keep_data=False):
        '''
        We call the function ```update``` every frame to capture an action from a skeleton, 
        and then generate controls.
        '''

        if self.tss and self.tss[-1] == ts:
            return

        # capture the skeleton from an image
        skeleton = self.holistic_tracker(image, ts)
        
        if keep_data:
            control = None
            self.df_data = df_data
        else:
            # skeleton -> action detection -> control
            control = self.module(skeleton)
        
        self.images.append(image)
        self.tss.append(ts)
        self.controls.append(control)

        self.buffer_resize()

    def release_keys(self):
        """
        Release keys which aren't pressed.
        """
        for action, control in self.module.mapping.items():
            if not isinstance(self.controls[-1], KeyControl):
                if isinstance(control, KeyControl):
                    control.release()
            else:
                if isinstance(control, KeyControl) and control.key != self.controls[-1].key:
                    control.release()

    def execute(self):
        """
        Execute controls (e.g. move mouse or press keys)
        """
        if len(self.controls) == 0:
            return 
        
        if not self.controls[-1]:
            return

        self.release_keys() # release not pressed keys

        if isinstance(self.controls[-1], Module):
            self.module = self.controls[-1]
        else:
            if self.controls[-1].execute:
                self.controls[-1].execute()
        
    def save(self, csv_path, image_dir, label):
        """
        For collecting data.
        """

        csv_dir = os.path.dirname(csv_path)
        image_name = str(int(time.time() * 1000)) + ".png"
        image_path = os.path.join(image_dir, image_name)

        pathlib.Path(csv_dir).mkdir(exist_ok=True, parents=True)
        pathlib.Path(image_dir).mkdir(exist_ok=True, parents=True)

        if label is None:
            label = input("Label: ")
            label = re.findall("\d+", label)[0]

        self.df_data["image_name"] = [image_name]
        self.df_data["label"] = [label]

        image = self.images[-1]

        if os.path.exists(csv_path):
            with open(csv_path, 'a') as f:
                write = csv.writer(f)
                write.writerow(self.df_data.values[0,:].tolist())
        else:
            df_data = self.df_data
            df_data.to_csv(csv_path, index=False)
        
        self.saved_images.append((image, image_path))

    def save_images(self):
        """
        Call save_images after the end of data collection, to save all images once.
        """

        for image, path in self.saved_images:
            image_pil = Image.fromarray(image)
            image_pil.save(path)


import os
import pathlib
import pickle
import torch
from ..utils.key_points import W_LIST_POSE, W2I_POSE, W_LIST_LEFT_HAND, W2I_LEFT_HAND, W_LIST_RIGHT_HAND, W2I_RIGHT_HAND
from ..utils.normalizer import normalize_data
from ..model import HACModel

class ActionDetector:
    """
    init: 
        model: trained model 
        pred2str: a function convert predictions to string
    """
    num_vertices = 33 + 21 * 2
    in_channels = 3

    def __init__(self, model):
        
        self.model = model
        self.target_columns = [key + "_x" for key in W_LIST_RIGHT_HAND] + [key + "_y" for key in W_LIST_RIGHT_HAND]
        self.target_columns += [key + "_x" for key in W_LIST_LEFT_HAND] + [key + "_y" for key in W_LIST_LEFT_HAND]
        self.target_columns += [key + "_x" for key in W_LIST_POSE] + [key + "_y" for key in W_LIST_POSE]

    def __call__(self, df):
        """
        input:
            x: inputs for model to predict hand gesture, the type of x depends on the model.
        output:
            output: hand gesture like "one", "two", "three", "stone"...
        """
        data = self.normalize(df)
        x = torch.from_numpy(data).view(-1, self.in_channels, self.num_vertices, 1).float().to(self.device)
        output = self.model(x) # batch, (x, y), num of joints, t
        pred = output.reshape(-1, len(self.mapping)).cpu().argmax(axis=1, keepdim=True)
        pred_action = self.pred2str(pred)
        print(pred_action)

        return pred_action

    def pred2str(self, pred):
        """
        input:
            pred: prediction result from the model
        output:
            output: pose like "stand", "jump"...
        """
        pass
    
    def normalize(self, df):
        df = normalize_data(df)
        return df.values

    '''
    def normalize(self, df):
        df = df.copy()
        df = df[self.target_columns].fillna(0)
        df = df[self.target_columns].dropna()
        df = df.apply(lambda x: (x - x.min()) / (x.max() - x.min()), axis=1)
        df = df.fillna(0)
        return df.values
    '''

class RobloxLiftGameActionDetector(ActionDetector):

    mapping = ["walk", "jump", "hands_on_hips", "point_left", "point_right", "arms_lift", "punch", "trample", "lateral_raise", "stand"]
    
    def __init__(self):

        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        num_class = len(self.mapping)
        model = HACModel(in_channels=3, num_class=num_class, k_hop=2, mode="pose_hand").to(self.device)
        model_path = os.path.join(pathlib.Path(__file__).parent.resolve(), "..\\trained_model\\gcn\\actions\\best_model.pth")
        model.load_state_dict(torch.load(model_path, map_location=torch.device(self.device)))
        model.eval()
        super(RobloxLiftGameActionDetector, self).__init__(model)
    
    def pred2str(self, pred):
        
        return self.mapping[pred[0]]


import os 
import pathlib
import pickle
import torch
from ..utils.key_points import W_LIST_POSE, W2I_POSE, W_LIST_LEFT_HAND, W2I_LEFT_HAND, W_LIST_RIGHT_HAND, W2I_RIGHT_HAND
from ..utils.normalizer import normalize_hand_data
from ..model import HACModel

class GestureDetector:

    num_vertices = 21 * 2
    in_channels = 3

    def __init__(self, model):

        """
        init: 
            model: trained model 
            pred2str: a function convert predictions to string
        """
        self.model = model
        self.target_columns = [key + "_x" for key in W_LIST_RIGHT_HAND] + [key + "_y" for key in W_LIST_RIGHT_HAND]
        self.target_columns += [key + "_x" for key in W_LIST_LEFT_HAND] + [key + "_y" for key in W_LIST_LEFT_HAND]
        #self.target_columns += [key + "_x" for key in W2I_POSE] + [key + "_y" for key in W2I_POSE]

    def __call__(self, df):
        """
        input:
            x: inputs for model to predict hand gesture, the type of x depends on the model.
        output:
            output: hand gesture like "one", "two", "three", "stone"...

        """
        data = self.normalize(df)
        x = torch.from_numpy(data).view(-1, self.in_channels, self.num_vertices, 1).float().to(self.device)
        output = self.model(x) # batch, (x, y), num of joints, t
        pred = output.reshape(-1, len(self.mapping)).cpu().argmax(axis=1, keepdim=True)
        pred_gesture = self.pred2str(pred)
        print(pred_gesture)

        return pred_gesture

    def pred2str(self, pred):
        """
        input:
            pred: prediction result from the model
        output:
            output: hand gesture like "one", "two", "three", "stone"...
        """
        pass

    def normalize(self, df):
        df = normalize_hand_data(df)
        return df.values

class MouseControlGestureDetector(GestureDetector):

    mapping = ["r_five", "r_zero", "l_five", "l_zero", "two_index_fingers_up", "two_index_fingers_down", "33", "55", "sit"]

    def __init__(self):
        
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        num_class = len(self.mapping)
        model = HACModel(in_channels=3, num_class=num_class, k_hop=2, mode="hand").to(self.device)
        model_path = os.path.join(pathlib.Path(__file__).parent.resolve(), "..\\trained_model\\gcn\\gestures\\best_model.pth")
        model.load_state_dict(torch.load(model_path, map_location=torch.device(self.device)))
        model.eval()
        super(MouseControlGestureDetector, self).__init__(model)

    def pred2str(self, pred):

        return self.mapping[pred[0]]


import torch
import torch.nn as nn
import torch.nn.functional as F
from .graph import Graph
from .gcn import GCN

class HACModel(nn.Module):
    def __init__(self, in_channels, num_class, k_hop, mode):
        super().__init__()
        
        self.in_channels = in_channels
        self.graph = Graph(k_hop=k_hop, mode=mode)
        A_k = torch.tensor(self.graph.A_k,
                           dtype=torch.float32,
                           requires_grad=False)
        self.register_buffer('A_k', A_k)
        
        self.data_bn = nn.BatchNorm2d(in_channels)
        self.gcns = nn.ModuleList((
            GCN(in_channels, 16, k_hop),
            GCN(16, 32, k_hop),
            GCN(32, 64, k_hop),           
        ))
        
        self.edge_importances = nn.ParameterList([
            nn.Parameter(torch.ones(self.A_k.size()))
            for i in self.gcns
        ])
        
        self.fcn = nn.Conv2d(64, num_class, kernel_size=1)
        self.drop_layer = nn.Dropout(p=0.5)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.data_bn(x)
        for gcn, importance in zip(self.gcns, self.edge_importances):
            x = gcn(x, self.A_k * importance)
        x = F.avg_pool2d(x, x.size()[2:])
        x = self.drop_layer(x)
        x = self.fcn(x).squeeze()
        #x = self.relu(x)
        return x


W_LIST_POSE = [
        "n", 
        "lei", 
        "le", 
        "leo", 
        "rei",
        "re",
        "reo",
        "lea",
        "rea",
        "ml",
        "mr",
        "ls",
        "rs",
        "lel",
        "rel",
        "lw",
        "rw",
        "lp",
        "rp",
        "li",
        "ri",
        "lt",
        "rt",
        "lh",
        "rh",
        "lk",
        "rk",
        "la",
        "ra",
        "lhe",
        "rhe",
        "lf",
        "rf"
    ]

W2I_POSE = {v: k for k, v in enumerate(W_LIST_POSE)}

W_LIST_LEFT_HAND = [
        "l_w",
        "l_t_c",
        "l_t_m",
        "l_t_i",
        "l_t_t",
        "l_i_m",
        "l_i_p",
        "l_i_d",
        "l_i_t",
        "l_m_m",
        "l_m_p",
        "l_m_d",
        "l_m_t",
        "l_r_m",
        "l_r_p",
        "l_r_d",
        "l_r_t",
        "l_p_m",
        "l_p_p",
        "l_p_d",
        "l_p_t",
    ]

W2I_LEFT_HAND = {v: k for k, v in enumerate(W_LIST_LEFT_HAND)}

W_LIST_RIGHT_HAND = [
    "r_w",
    "r_t_c",
    "r_t_m",
    "r_t_i",
    "r_t_t",
    "r_i_m",
    "r_i_p",
    "r_i_d",
    "r_i_t",
    "r_m_m",
    "r_m_p",
    "r_m_d",
    "r_m_t",
    "r_r_m",
    "r_r_p",
    "r_r_d",
    "r_r_t",
    "r_p_m",
    "r_p_p",
    "r_p_d",
    "r_p_t",
]

W2I_RIGHT_HAND = {v: k for k, v in enumerate(W_LIST_RIGHT_HAND)}

ALL_COLUMNS = W_LIST_POSE + W_LIST_LEFT_HAND + W_LIST_RIGHT_HAND
X_COLS = []
Y_COLS = []
V_COLS = []
for c in ALL_COLUMNS:
    X_COLS.append(c + "_x")
    Y_COLS.append(c + "_y")
    V_COLS.append(c + "_v")
    
LEFT_HAND_X_COLS = []
LEFT_HAND_Y_COLS = []
LEFT_HAND_V_COLS = []
for c in W_LIST_LEFT_HAND:
    LEFT_HAND_X_COLS.append(c + "_x")
    LEFT_HAND_Y_COLS.append(c + "_y")
    LEFT_HAND_V_COLS.append(c + "_v")
    
RIGHT_HAND_X_COLS = []
RIGHT_HAND_Y_COLS = []
RIGHT_HAND_V_COLS = []
for c in W_LIST_RIGHT_HAND:
    RIGHT_HAND_X_COLS.append(c + "_x")
    RIGHT_HAND_Y_COLS.append(c + "_y")
    RIGHT_HAND_V_COLS.append(c + "_v")
    
HAND_XY_COLS = LEFT_HAND_X_COLS + RIGHT_HAND_X_COLS + LEFT_HAND_Y_COLS + RIGHT_HAND_Y_COLS + LEFT_HAND_V_COLS + RIGHT_HAND_V_COLS
ALL_XY_COLS = X_COLS + Y_COLS + V_COLS


import mediapipe as mp
import pandas as pd
from google.protobuf.json_format import MessageToDict
from ..utils.key_points import W_LIST_POSE, W2I_POSE, W_LIST_LEFT_HAND, W2I_LEFT_HAND, W_LIST_RIGHT_HAND, W2I_RIGHT_HAND

class HandTracker:
    
    def __init__(self):
        self.mp_hands = mp.solutions.hands
        self.hands = self.mp_hands.Hands(min_detection_confidence=0.3,
                               min_tracking_confidence=0.3, max_num_hands=2, static_image_mode=False)
        self.mp_drawing = mp.solutions.drawing_utils
        self.result_hands = None
        self.target_columns = [key + "_x" for key in W_LIST_RIGHT_HAND] + [key + "_y" for key in W_LIST_RIGHT_HAND]
        self.target_columns += [key + "_x" for key in W_LIST_LEFT_HAND] + [key + "_y" for key in W_LIST_LEFT_HAND]

    def __call__(self, image, ts):
        """
        input:
            image: image
        output:
            output: hand_skeleton
        """

        self.results_hands = self.hands.process(image)
        keypoints_dict = self.pred_to_dict(self.results_hands, ts)
        hand_skeleton = self.dict_to_data(keypoints_dict)

        return hand_skeleton

    def dict_to_data(self, keypoints_dict):
        df = pd.DataFrame(keypoints_dict, index=[0])               

        return df

    def pred_to_dict(self, results_hands, ts):
        
        cp_dict = {}
        cp_dict_list = []

        best_left_idx = -1
        best_left_score = 0.0
        best_right_idx = -1
        best_right_score = 0.0

        if results_hands and results_hands.multi_hand_landmarks:

            for idx, hand_landmarks in enumerate(results_hands.multi_handedness):
                
                classification = MessageToDict(results_hands.multi_handedness[idx].classification[0])

                if classification['score'] < 0.9:
                    continue

                if classification['label'] == "Left" and classification['score'] > best_left_score:
                    best_left_idx = idx
                    best_left_score = classification['score']

                if classification['label'] == "Right" and classification['score'] > best_right_score:
                    best_right_idx = idx
                    best_right_score = classification['score']

        if best_left_idx > -1:
            cp_dict_list.append(
                self.landmarks_to_dict(results_hands.multi_hand_landmarks[best_left_idx].landmark, W2I_LEFT_HAND, ts)
            )
        else:
            cp_dict_list.append(
                self.landmarks_to_dict_dummy(W2I_LEFT_HAND, ts)
            )
        
        if best_right_idx > -1:
            cp_dict_list.append(
                self.landmarks_to_dict(results_hands.multi_hand_landmarks[best_right_idx].landmark, W2I_RIGHT_HAND, ts)
            )
        else:
            cp_dict_list.append(
                self.landmarks_to_dict_dummy(W2I_RIGHT_HAND, ts)
            )
        
        for cp in cp_dict_list:
            cp_dict.update(cp)

        return cp_dict

    def landmarks_to_dict_dummy(self, w2i, ts):
 
        cp_dict = {}

        for w in w2i:
            cp_dict[w + "_x"] = 0.0
            cp_dict[w + "_y"] = 0.0
            cp_dict[w + "_v"] = 0.0

        cp_dict["ts"] = ts
        
        return cp_dict

    def landmarks_to_dict(self, landmarks, w2i, ts):

        cp_dict = {}

        for w in w2i:
            cp_dict[w + "_x"] = landmarks[w2i[w]].x
            cp_dict[w + "_y"] = -landmarks[w2i[w]].y + 1
            cp_dict[w + "_v"] = landmarks[w2i[w]].visibility

        cp_dict["ts"] = ts
        
        return cp_dict

    def draw_landmarks(self, image):

        if self.results_hands and self.results_hands.multi_hand_landmarks:
            for idx, hand_landmarks in enumerate(self.results_hands.multi_hand_landmarks):
                self.mp_drawing.draw_landmarks(
                        image, hand_landmarks, self.mp_hands.HAND_CONNECTIONS)

        return image


import torch
import torch.nn as nn
import torch.nn.functional as F

class GCN(nn.Module):
    def __init__(self, in_channels, out_channels, k_hop):
        super().__init__()
        
        self.conv = nn.Conv2d(in_channels,
                              out_channels * k_hop,
                              kernel_size=(1, 1),
                              padding=(0, 0),
                              stride=(1, 1))
        
        if in_channels == out_channels:
            self.residual = lambda x: x

        else:
            self.residual = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=1,
                    stride=(1, 1)),
                nn.BatchNorm2d(out_channels),
            )
        
    def forward(self, x, A_k):
        
        res = self.residual(x)
        
        B, F, J, T = x.size()
        
        K = A_k.size()[0]
        
        x = self.conv(x)
        B, KC, J, T = x.size()
        
        x = x.view(B, K, KC // K, J, T)
        output = torch.einsum('bkcjt,kji->bcit', (x, A_k)).contiguous() + res
        
        return output


import pydirectinput
import mouse
import threading
import time
import keyboard

pydirectinput.PAUSE = 0

class MouseControl:

    freeze = False

    def __init__(self, method, **params):
        self.execute = getattr(self, method)
        self.method_name = method

        for key in params:
            setattr(self, key, params[key])
    
    def _check_freeze(func):
        def wrap(self, **params):
            sec = 0.01
            if not self.freeze:
                self.freeze = True
                func(self, **params)
                t = threading.Thread(target=self.freeze_timer, daemon=True, args=(sec,))
                t.start()

        return wrap

    @_check_freeze
    def click(self):
        mouse.click()

    @_check_freeze
    def mouse_left_down(self):
        mouse.press()

    @_check_freeze
    def mouse_left_up(self):
        mouse.release()

    @_check_freeze
    def mouse_right_down(self):
        mouse.press(button="right")

    @_check_freeze
    def mouse_right_up(self):
        mouse.release(button="right")

    @_check_freeze
    def right_click(self):
        mouse.right_click()

    @_check_freeze
    def roll_up(self):
        mouse.wheel(30) 
        print("roll_up")

    @_check_freeze
    def roll_down(self):
        mouse.wheel(-30) 
        print("roll_down")

    @_check_freeze
    def move_diff(self, factor=4000):
        print("move_diff")
        x1 = self.df_data_1_x.values.mean()
        x2 = self.df_data_2_x.values.mean()
        y1 = self.df_data_1_y.values.mean()
        y2 = self.df_data_2_y.values.mean()
        dx = -(x2 - x1).item() * factor
        dy = -(y2 - y1).item() * factor
        pydirectinput.move(int(dx), int(dy))
    
    def right_move_diff(self):
        self.move_diff()

    def left_move_diff(self):
        self.move_diff()

    def set_params(self, **params):
        for key in params:
            setattr(self, key, params[key]) 

    def freeze_timer(self, sec):
        time.sleep(sec)
        self.freeze = False

class KeyControl:
    def __init__(self, key, **params):
        self.key = key
        self.method_name = key

    def execute(self):

        if self.key == "skip":
            return

        if self.key == "left":
            pydirectinput.keyDown('left')
        elif self.key == "right":
            pydirectinput.keyDown('right')
        elif self.key == "skip":
            pass
        else:
            keyboard.press(self.key)

    def release(self):

        if self.key == "skip":
            return

        keyboard.release(self.key)




import pandas
from .key_points import X_COLS, Y_COLS, \
                        ALL_XY_COLS, HAND_XY_COLS, \
                        LEFT_HAND_X_COLS, LEFT_HAND_Y_COLS, \
                        RIGHT_HAND_X_COLS, RIGHT_HAND_Y_COLS

def normalize_data(df):
    
    df_data = df[ALL_XY_COLS].copy()
    df_data = df_data.fillna(0)
    
    center_x_cols = ['lh_x', 'rh_x']
    center_y_cols = ['lh_y', 'rh_y']
    
    x_center = df[center_x_cols].mean(axis=1)
    y_center = df[center_y_cols].mean(axis=1)
    
    df_data[X_COLS] = df[X_COLS].sub(x_center, axis=0)
    df_data[Y_COLS] = df[Y_COLS].sub(y_center, axis=0)
    
    x_width = (df_data[X_COLS].max(axis=1) - df_data[X_COLS].min(axis=1) + 1e-9)
    df_data[X_COLS] = df_data[X_COLS].div(x_width, axis=0)
    df_data[Y_COLS] = df_data[Y_COLS].div(x_width, axis=0) # ratio resize

    return df_data
    
def normalize_hand_data(df):
    
    df_hand_data = df[HAND_XY_COLS].copy()
    df_hand_data = df_hand_data.fillna(0)
    
    left_hand_center_x_cols = ['l_w_x']
    left_hand_center_y_cols = ['l_w_y']
    right_hand_center_x_cols = ['r_w_x']
    right_hand_center_y_cols = ['r_w_y']
    
    x_left_hand_root = df_hand_data[left_hand_center_x_cols].mean(axis=1)
    y_left_hand_root = df_hand_data[left_hand_center_y_cols].mean(axis=1)
    x_right_hand_root = df_hand_data[right_hand_center_x_cols].mean(axis=1)
    y_right_hand_root = df_hand_data[right_hand_center_y_cols].mean(axis=1)
    
    df_hand_data[LEFT_HAND_X_COLS] = df[LEFT_HAND_X_COLS].sub(x_left_hand_root, axis=0)
    df_hand_data[LEFT_HAND_Y_COLS] = df[LEFT_HAND_Y_COLS].sub(y_left_hand_root, axis=0)
    df_hand_data[RIGHT_HAND_X_COLS] = df[RIGHT_HAND_X_COLS].sub(x_right_hand_root, axis=0)
    df_hand_data[RIGHT_HAND_Y_COLS] = df[RIGHT_HAND_Y_COLS].sub(y_right_hand_root, axis=0)
    
    left_x_width = (df_hand_data[LEFT_HAND_X_COLS].max(axis=1) - \
                    df_hand_data[LEFT_HAND_X_COLS].min(axis=1) + 1e-9)
    right_x_width = (df_hand_data[RIGHT_HAND_X_COLS].max(axis=1) - \
                     df_hand_data[RIGHT_HAND_X_COLS].min(axis=1) + 1e-9)
    
    df_hand_data[LEFT_HAND_X_COLS] = df_hand_data[LEFT_HAND_X_COLS].div(left_x_width, axis=0)
    df_hand_data[LEFT_HAND_Y_COLS] = df_hand_data[LEFT_HAND_Y_COLS].div(left_x_width, axis=0)
    df_hand_data[RIGHT_HAND_X_COLS] = df_hand_data[RIGHT_HAND_X_COLS].div(right_x_width, axis=0)
    df_hand_data[RIGHT_HAND_Y_COLS] = df_hand_data[RIGHT_HAND_Y_COLS].div(right_x_width, axis=0)
    
    return df_hand_data